// generate .mc file
// $Id: generate.cc,v 1.3 2000/02/05 00:41:27 waider Exp $
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <limits.h>
#include <misc.h>
#include "internal.h"
#include "paths.h"
#include <daemoni.h>
#include <netconf_def.h>
#include <netconf.pe>
#include "thishost.h"
#include "m4mailconf.h"
#include "m4mailconf.m"
#include <popen.h>

extern M4MAILCONF_HELP_FILE help_m4mailconf;

CONFIG_FILE f_sendmail_mc (ETC_SENDMAIL_MC,help_m4mailconf
	,CONFIGF_GENERATED|CONFIGF_OPTIONNAL|CONFIGF_PROBED
	,subsys_sendmail);

CONFIG_FILE f_virtuser (ETC_MAIL_VIRTUSER,help_m4mailconf
	,CONFIGF_OPTIONNAL|CONFIGF_PROBED
	,subsys_mail);
CONFIG_FILE f_pophash (ETC_MAIL_POPHASH,help_m4mailconf
	,CONFIGF_OPTIONNAL|CONFIGF_PROBED
	,subsys_mail);

extern CONFIG_FILE f_mailtable;
extern CONFIG_FILE f_sendmail_cw;

const int TOKEN_VALLEN = 100;

/*
	Produce the file /etc/sendmail.mc using the configuration.

	It may generate the normal sendmail.mc or a temporary one
	for automated test purpose (If its argument is true).
	Not implemented yet.

	Return -1 if any error
*/
PRIVATE int M4MAILCONF::generate(FILE *fout)
{
	int	ret = 0;
	THISHOST thishost;

	// Things we will need
	const char *hname = thishost.getname1();


	// CHROME: VERSIONID
	// Allows you to RCS-tag your sendmail revisions
	// FIXME allow the user to customize this, particularly the
	// "8.8 (Berkeley)" bit
	fputs( "VERSIONID(`@(#) sendmail.mc 8.8 (Berkeley) $Revision: 1.3 $')dnl\n",
		   fout );

	// REQUIRED: OSTYPE
	// If linuxconf migrates to other planets^H^H^H^H^H^H^H*nix
	// variants, we'll need to change this.
	fputs( "OSTYPE(`linux')dnl\n", fout );

	/* #specification: generating sendmail.mc / other module contribution
		When generating the sendmail.mc, the m4mailconf module allows
		other modules to participate in the generation. We are using the
		general purpose messaging facility. Various messages are sent
		at different stage of the sendmail.mc generation.

		The messaging api is text only using an argc,argv[] pair. We
		encode a FILE pointer as the first argument. A strategy used by
		co-manager might be better.
	*/
	const char *message_argv[2];
	message_argv[0] = (const char *)fout;
	message_argv[1] = NULL;
	module_sendmessage ("m4mailconf:defclass",1,message_argv);

	// Aliases for this host
	// This isn't modifying the .mc file, so arguably we could migrate
	// it elsewhere i.e. the confread.cc file in the write() method.

	// The write() method is used to handle the case where mainserv is
	// unset, anyway. Sorry, but I'm lazy.

	// For lack of anything better to do, we'll drop most of this in
	// as-is. No harm done, mostly, unless you've got a hostname like
	// "define.foo.bar" which m4 may pick up and eat because it's
	// GREEDY. Bad m4!
	const char *dom = thishost.getdomain();
	if (mainserv) {
	  if (dom != NULL) {
		// check if our domain is in the alias list. If it ain't, add it.
		if ( !alias.lookup( dom )) {
		  alias.add( new SSTRING( dom ));
		}
	  }else{
		xconf_error (MSG_U(E_NODOMAIN,
						   "Can't compute the domain name of this server\n"
						   "It won't accept email correctly.\n"
						   "Please enter a fully qualified domain name in the\n"
						   "\"Basic host information\" screen."));
	  }
	}

	// We'll try and keep the "define..." lines together, and sort
	// them by conf... vs non-conf... lines (e.g. SMARTHOST, MAIL_HUB)
	fputs( "\n", fout );

	// EXPOSED_USER
	// not a define, so sue me. sendmail is wacky, even using the m4
	// config.
	// The doco isn't clear on whether we should do these a line at a
	// time or all in one go. I've taken the former option until I
	// figure otherwise.
	{
	  const char *pt = users.dontmasque.get();
	  while( 1 ){ //
		char word[1025];
		pt = str_copyword( word, pt, sizeof(word)-1);
		if (word[0] == '\0' ) break;
		fprintf( fout, "EXPOSED_USER(%s)dnl\n", word );
	  }
	}

	// Local users
	{
		const char *pt = users.deliverlocal.get();
		while (1){
			char word[1025];
			pt=str_copyword(word,pt,sizeof(word)-1);
			if (word[0] == '\0') break;
			fprintf (fout,"LOCAL_USER(`%s')dnl\n",word);
		}
	}

	// Mail hub
	if ((mailhost.icmp(hname)!=0) && (!mailhost.is_empty())) {
		fprintf (fout,"MAIL_HUB(`%s')dnl\n", mailhost.get());
	}

	if (!mailas.is_empty())	
	  fprintf (fout,"MASQUERADE_AS(%s)dnl\n",mailas.get());

	fputs( "\n", fout );

	// Smarthost
	/* #Specification: m4mailconf / sendmail.mc / smarthost
		If we specify that we are our own smarthost, this
		causes a problem with sendmail. It could be
		argued that it makes sense (independant of sendmail).

		Given also that linuxconf mail configuration
		will be shareable one day, this makes sense.

		When we detect that the smarthost is this host
		we simply ignore it (The S macro of sendmail.mc
		remains empty).

		We do the same for the mailhost (Macro H of
		sendmail.mc)
	*/
	if ((smarthost.icmp(hname)!=0) && (!smarthost.is_empty())){
		fprintf( fout, "define(`SMARTHOST',`%s:%s')\n",
				 smartmailer.get(), smarthost.get());
	}

	// UUCP Max message size
	if (features.uucpmax > 0){
	  fprintf( fout, "define(`UUCP_MAILER_MAX', `%d')\n",
			   features.uucpmax );
	}

	// UUCP nobatch
	// NB this uses the default UUCP_MAILER_ARGS as specified in
	// 8.9.3, except that it switches -r on or off as apropriate. This
	// could do with being checked for newer/older sendmail versions.
	// FIXME If I was writing this in Perl, I may well have done some
	// horrific regexp evil here.
	fprintf( fout, 
			 "define(`UUCP_MAILER_ARGS', `uux -%s -a$g -gC $h!rmail ($u)')\n",
			 features.uucpnobatch ? "" : " -r" );


	// User/Group and mail spool specification
	{
		char userrepl[2][TOKEN_VALLEN];
		strcpy (userrepl[0],"mail:mail");
		if (strcmp(linuxconf_getdistdir(),"suse")==0){
			strcpy (userrepl[0],"bin:bin");
		}
		strcpy (userrepl[1],configf_lookuppath("/var/spool/mqueue"));

		// NB THIS LINE SEPARATES define(`FOO... from
		// define(`confFOO..., hence the double "\n". This is an
		// aesthetic conceit.
		fprintf( fout, "define(`QUEUE_DIR', `%s')\n\n", userrepl[1] );
		fprintf( fout, "define(`confDEF_USER_ID', `%s')\n",
				 userrepl[0] );
	}

	// Trusted users
	// FIXME should check if there are any!
	fprintf( fout, "define(`confTRUSTED_USERS', `%s')\n", users.trust.get());

	// GECOS
	if ( features.usegecos )
	  fputs( "define(`confMATCH_GECOS', `True')\n", fout );

	// Max message size
	if ( features.maxmsgsize > 0 )
	  fprintf( fout, "define(`confMAX_MESSAGE_SIZE', `%d')\n",
			   features.maxmsgsize );

	// Max recipients per message
	if ( features.maxrecipients > 0 )
	  fprintf( fout, "define(`confMAX_RCPTS_PER_MESSAGE', `%d')\n",
			   features.maxrecipients );

	// Delivery mode
	if ( features.deferdeliv )
	  fputs( "define(`confDELIVERY_MODE', `deferred')\n", fout );

	// Bogus HELO
	if ( features.bogushelo )
	  fputs( "define(`confALLOW_BOGUS_HELO', `True')\n", fout );

	// Ideally we would remove the domain from the alias list here,
	// but, eh, SSTRING class doesn't seem to have any remove( item )
	// method. Bah.
	
	// Stick all the FEATUREs into a block
	fputs( "\n", fout );

	fprintf (fout,"Cw %s\n",hname); // FIXME ditto above FIXME
	if (sendmail_cw){
	  fprintf (fout,
			   "define(`confCW_FILE', `-o %s')\nFEATURE(`use_cw_file')dnl\n",
			   f_sendmail_cw.getpath());
	}else{
	  for (int i=0; i<alias.getnb(); i++){
		fprintf (fout,"Cw %s\n",alias.getitem(i)->get());
	  }
	}

	// smrsh
	{
		char progrepl[1][TOKEN_VALLEN];
		strcpy (progrepl[0],"/bin/sh");
		if (features.usesmrsh){
			DAEMON_INTERNAL *dae = daemon_find ("smrsh");
			if (dae != NULL){
				const char *path = dae->getpath();
				if (file_exist (path)){
				  strcpy (progrepl[0],path);
				}else if (!simul_ison()){
				  xconf_error (MSG_U(E_NOSMRSH, 
									 "The smrsh shell is not installed\n"
									 "on this system.\n"
									 "Generating a sendmail.mc using /bin/sh"));
				}
			}
		}
		fprintf( fout, "FEATURE(`smrsh',`%s')dnl\n", progrepl[ 0 ] );
	}

	// mailertable
	// special (domain) routing / special routing database
	// NB the database format selector controls ALL databases, not
	// just this one. We'll call this a feature for now.
	if (features.mailertable){
	  fprintf( fout, "FEATURE(`mailertable',`%s -o %s')dnl\n",
			   features.dbformat.get(), f_mailtable.getpath());
	}else{
	  if (spcs.getnb()>0){
		xconf_notice(MSG_U(E_MTABLEON,
						   "The /etc/sendmail.mc will be generated without\n"
						   "special (domain) routing enabled. This system\n"
						   "has already some special routing definition(s)\n"
						   "and this sendmail.mc won't make use of them.\n"
						   "Enable \"special routing database\" in the\n"
						   "\"basic information\" dialog to fix this\n"
						   "and regenerate the sendmail.mc."));
	  }
	}
	  
	// virtusertable
	// handles delivery rerouting w/o changing mail headers
	// i.e. complex (user) routing
	// The logic is a bit bogus. We don't need this feature if the
	// file exists but is empty, for example. FITNR.
	if (f_virtuser.exist()){
	  fprintf (fout,"FEATURE(`virtusertable', `%s -o %s')dnl\n",
			   features.dbformat.get() ,f_virtuser.getpath());
	}

	// relay control
	// if relayctrl is unset, we ignore all the other spam control
	// settings, cos heck, if you don't want spam control, we're not
	// gonna give it to you, right?
	if ( features.relayctrl ){

	  // Individual spam control features. Order per the README file.
	  if ( features.relay_hostsonly ) {
		fputs( "FEATURE(relay_hosts_only)dnl\n", fout );
	  }

	  if ( features.relay_domain ) {
		fputs( "FEATURE(relay_domain)dnl\n", fout );
	  }

	  if ( features.relay_accessdb ) {
		fputs( "FEATURE(access_db)dnl\n", fout );
	  }

	  if ( features.relay_blacklist ) {
		// this MUST have access_db, sorry bub.
		if ( !features.relay_accessdb ) {
		  // Whoops. No accessdb => no blacklist_recipients. Sorry.
		  xconf_notice(MSG_U(E_ACCESSDBON,
							 "You need to enable accessdb in order to use"
							 "the recipients blacklist feature."));
		} else {
		  fputs( "FEATURE(blacklist_recipients)dnl\n", fout );
		}
	  }

	  if ( features.relay_rbl ) {
		fputs( "FEATURE(rbl)dnl\n", fout );
	  }

	  if ( features.relay_unqual ) {
		fputs( "FEATURE(accept_unqualified_senders)dnl\n", fout );
	  }

	  if ( features.relay_unresolv ) {
		fputs( "FEATURE(accept_unresolvable_domains)dnl\n", fout );
	  }

	  if ( features.relay_mx ) {
		fputs( "FEATURE(relay_based_on_mx)dnl\n", fout );
	  }

	  // BAD IDEAS
	  // Using these is probably the wrong thing, but I'm not going to
	  // cast asparagus on you for that.
	  if ( features.relay_local_from ) {
		fputs( "FEATURE(relay_local_from)dnl\n", fout );
	  }

	  if ( features.relay_loose ) {
		fputs( "FEATURE(loose_relay_check)dnl\n", fout );
	  }
	} else {
	  fputs( "FEATURE(`promiscuous_relay')dnl\n", fout );
	  
	  // We will, however, let you know you're doing it wrong.
	  xconf_notice(MSG_U(E_MSPAMMER,
						 "You are generating a sendmail configuration with\n"
						 "spam control disabled. If your host is connected\n"
						 "to the Internet, this is a VERY BAD IDEA.\n"
						 "Enable \"Enable relay control (spammers)\" in the\n"
						 "\"basic information\" dialog to fix this\n"
						 "and regenerate sendmail.mc."));
	}

	//  // BEGIN Virtual domains
//  	fputs ("# Virtual email domain\n",fout);
//  	VDOMAINS vdomains;		// Virtual domains
//  	for (int i=0; i<vdomains.getnb(); i++){
//  		VDOMAIN *v = vdomains.getitem(i);
//  		//fprintf (fout,"CV %s\n",v->domain.get());
//  		for (int o=0; o<v->others.getnb(); o++){
//  			SSTRING *s = v->others.getitem(o);
//  			//fprintf (fout,"CV %s\n",s->get());
//  		}
//  	}
	// END Virtual domains

	// BEGIN pophash
//  	if (f_pophash.exist()){
//  		fputs ("# Roaming pop users accces\n",fout);
//  		fprintf (fout,"Kpopauth %s %s\n",features.dbformat.get()
//  			,f_pophash.getpath());
//  	}
	// END pophash

//  	if (features.relayctrl){
//  		static const char *usertoken[]={
//  			"$(POPHASH)"
//  		};
//  		char userrepl[1][TOKEN_VALLEN];
//  		if (f_pophash.exist()){
//  			strcpy (userrepl[0],"");
//  		}else{
//  			strcpy (userrepl[0],"#");
//  		}
//  	}

//  	if (f_spam_deny.exist()){
//  	}else{
//  		fputs ("#Empty junk rule\nSjunk\n\n",fout);
//  	}

	// I'm not 100% happy with these, but what the heck.

	// Virtual domains **** UNFINISHED ****
	// This is just here so the complex rule stuff works.
	// fputs ("# Virtual email domain\n",fout);
	VDOMAINS vdomains;		// Virtual domains
	for (int i=0; i<vdomains.getnb(); i++){
		VDOMAIN *v = vdomains.getitem(i);
		//fprintf (fout,"CV %s\n",v->domain.get());
		for (int o=0; o<v->others.getnb(); o++){
			SSTRING *s = v->others.getitem(o);
			//fprintf (fout,"CV %s\n",s->get());
		}
	}

	// Let a module define complex/exceptional routing
  	module_sendmessage ("m4mailconf:complex_s0",1,message_argv);
  	COMPLEX_ROUTES cplx;
  	cplx.rule0 (fout,alias,vdomains);

	// masqs rules
	ret |= masqs.rule1(fout);

	// REQUIRED: MAILER
	// Visual separation
	fputs( "\n", fout );

	// First, get the local mailer path
	// Euw, this is horrible. We should maybe use PATH and stuff, or
	// try a range of directories including /usr/local/bin
	if (deliver.is_empty()){
	  if (file_exist ("/bin/mail.local")){
		fputs( "define(`LOCAL_MAILER_PATH', `/bin/mail.local')\n", fout );
	  }else if (file_exist ("/usr/bin/deliver")){
		fputs( "define(`LOCAL_MAILER_PATH', `/usr/bin/deliver')\n", fout );
	  }else if (file_exist ("/usr/bin/procmail")){
		fputs( "define(`LOCAL_MAILER_PATH', `/usr/bin/procmail')\n", fout );
	  }
	}else if (deliver.cmp("deliver")==0){
	  fputs( "define(`LOCAL_MAILER_PATH', `/bin/mail.local')\n", fout );
	}else if (deliver.cmp("procmail")==0){
	  fputs( "define(`LOCAL_MAILER_PATH', `/usr/bin/deliver')\n", fout );
	}else if (deliver.cmp("mail.local")==0){
	  fputs( "define(`LOCAL_MAILER_PATH', `/usr/bin/procmail')\n", fout );
	}
	
	fputs( "MAILER(local)dnl\n", fout );
	
	// FIXME not necessarily present. How can we determine if it's
	// necessary? 
	// * check for SMTP dependant hackery?
	fputs( "MAILER(smtp)dnl\n", fout );

	// Got fax?
	// FIXME I need to check how this sits with Jacques' faxstuff.
	if ( mfax.enable )
	  fputs( "MAILER(fax)dnl\n", fout );

	// Let modules record mailer
	module_sendmessage ("m4mailconf:defmailer",1,message_argv);

	return ret;
}
	
static const char K_SENDMAILMCSUM[]="sendmail_mc_sum";

/*
	Check if the sendmail.mc currently installed was generated by me
	Use the MD5 checksum to make sure.

	sendmail_sum[] will contain the MD5 checksum of the current sendmail.mc
*/
PRIVATE bool M4MAILCONF::generated_byme(char sendmail_sum[])
{
	bool ret = false;
	const char *gensum = confread_getval (K_SENDMAILMCSUM,"");
	if (f_sendmail_mc.md5sum(sendmail_sum)!= -1
		&& strcmp(gensum,sendmail_sum)==0){
		ret = true;
	}
	return ret;
}

PUBLIC int M4MAILCONF::generate_go()
{
	int ret = -1;

#ifdef DEBUG
	fprintf( stderr, "ready to generate_go\n" );
#endif

	FILE *fout = f_sendmail_mc.fopen ("w");
	if (fout != NULL){
	  ret = generate (fout);
	  if (features.mailertable) spcs.build();
	  fclose (fout);
	  char sum[100];
	  f_sendmail_mc.md5sum(sum);
	  linuxconf_setcursys(subsys_sendmail);
	  confread_replace (K_SENDMAILMCSUM,sum);
	  linuxconf_save();
	  if (ret != -1){
		xconf_notice (MSG_U(N_HASGEN,"%s has been regenerated!")
					  ,f_sendmail_mc.getpath());
	  }
	} else {
	  perror( "generate_go" );
	}
	return ret;
}

/*
	Produce the file /etc/sendmail.mc using the configuration.

	It may generate the normal sendmail.mc or a temporary one
	for automated test purpose (If its argument is true).
	Not implemented yet.

	Return -1 if any error
*/
PUBLIC int M4MAILCONF::generate()
{
	int ret = -1;
	const char *path = f_sendmail_mc.getpath();
	net_prtlog (NETLOG_CMD,MSG_U(I_UPDSENDMAILCF,
								 "Generating %s\n"),path);
	bool go = true;
	char sum[100];
	if (!generated_byme(sum)){
		char pathold[PATH_MAX];
		sprintf (pathold,"%s.old",path);
		char bufmsg[1000];
		snprintf (bufmsg,sizeof(bufmsg)-1,MSG_U(Q_NOTBYME,
			 "The file %s was not generated by Linuxconf.\n"
			 "It was either never generated by Linuxconf\n"
			 "or has been modified manually later.\n"
			 "\n"
			 "Linuxconf will overwrite it and produce a backup file\n"
			 "named %s.\n"
			 "\n"
			 "\tOk ?")
			,path,pathold);
		if (xconf_yesno (MSG_U(T_NOTBYME,"Attention"),bufmsg,help_nil)
			!=MENU_YES){
			go = false;
			net_prtlog (NETLOG_VERB,MSG_U(I_USERSAYNO
				,"Not done as confirmed by the admin\n"));
		}else{
			path = f_sendmail_mc.getpath();
			net_prtlog (NETLOG_CMD,"mv %s %s\n",path,pathold);
			rename (path,pathold);
		}
	}
	if (go) ret = generate_go();
	return ret;
}
/*
	Return true if the sendmail.mc was generated once by Linuxconf
*/
bool m4mailconf_generated_once ()
{
	// No need to check that the configuration was ever edited
	// by linuxconf. We simply check that linuxconf did generate at
	// least once a sendmail.mc. This proves the user has visited
	// the sendmail configuration menu
	const char *gensum = confread_getval (K_SENDMAILMCSUM,"");
	return strcmp(gensum,"")!=0;
}
/*
	Generate a new sendmail.mc if needed.
	We compare the md5 checksum of sendmail.mc with the one we would get
	if the file was generated from linuxconf config. If the checksum
	differ, then sendmail.mc needs to be updated.
	Return	1 if the sendmail.mc was regenerated
				(or should be if run in simul mode)
			0 if all is fine
			-1 if any error
*/
PUBLIC int M4MAILCONF::generate_if()
{
  int ret = -1;

  if (m4mailconf_generated_once()){
	POPEN pop ("md5sum","");
	if (pop.isok()){
	  // Imitate the way CONFIG_FILE::md5sum works by sending this line
	  SSTREAM_POPEN ss (pop);
	  configf_sendexist (ss,true);
	  FILE *fout = pop.getfout();
	  ret = generate (fout);
	  pop.close();
	  if (ret != -1){
		char line[100];
		if (pop.readout(line,sizeof(line)-1)==0){
		  char sum_file[100],sum_conf[100];
		  str_copyword (sum_conf,line);
		  if (generated_byme (sum_file) && strcmp(sum_conf,sum_file)!=0){
			net_prtlog (NETLOG_CMD,MSG_R(I_UPDSENDMAILCF),
						f_sendmail_mc.getpath());
			ret = 1;
			if (!simul_ison()){
			  ret = generate_go();
			  if (ret != -1) ret = 1;
			}
		  }
		}
	  }
	}
  }
  return ret;
}

